package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;

/**
 * Default implementation of {@link LightSource} interface.
 * Simple and supersampled variants point to each other.
 *
 * @version 0.26 $Rev: 465 $ $Date: 2007-05-25 23:57:10 +0200 (Fri, 25 May 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultLightSource.java">DefaultLightSource.java</a>
 */
public abstract class DefaultLightSource extends Piece implements LightSource
{
  //==========================================================================
  //  Data:

  /** JGL script key: lights. */
  public final static String KEY_LIGHT = "light";

  /** Property name: light source parameters. */
  public final static String PARAMS = "Params";

  /** JGL script file-name. */
  protected String scriptName;

  /** Object used for light-source visibility tests. */
  protected Intersectable inter = null;

  /** Accelerator: connected intersectable was tried? */
  protected boolean plugInter = false;

  //==========================================================================
  //  Support:

  /**
   * Asserts the {@link #inter} accelerator.
   *
   * @return {@code true} if failed.
   */
  protected boolean assertInter ()
  {
    if ( inter == null &&
         !plugInter )
    {
      inter     = (Intersectable)getInterface( PL_OUTPUT, IFACE+"Intersectable" );
      plugInter = true;
    }

    return( inter == null );
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Sets intersectable object (for light-source of visibility: {@link #lights(MicroFacet)}, ..).
   */
  public Intersectable setIntersectable ( Intersectable intersect )
  {
    plugInter = false;
    Intersectable old = inter;
    inter = intersect;
    return old;
  }

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return <tt>true</tt> if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  public boolean lights ( MicroFacet mf )
  {
    return lights( mf, 0, 1 );
  }

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   * Super-sampling variant.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @return <tt>true</tt> if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  public boolean lights ( MicroFacet mf, int order, int total )
  {
    return lights( mf );
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return <tt>true</tt> if the light source can light on the given point (<b>cannot</b>
   *         replace the <tt>lights()</tt> call!).
   * @see    #lights
   * @see    #getDirection
   */
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    return getIntensity( mf, 0, 1, color );
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   * Super-sampling variant.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return <tt>true</tt> if the light source can light on the given point (<b>cannot</b>
   *         replace the <tt>lights()</tt> call!).
   * @see    #lights
   * @see    #getDirection
   */
  public boolean getIntensity ( MicroFacet mf, int order, int total, double[] color )
  {
    return getIntensity( mf, color );
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return <tt>false</tt> if the light source is not directional (thus <tt>dir</tt>
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    return getDirection( mf, 0, 1, dir );
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   * Super-sampling variant.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return <tt>false</tt> if the light source is not directional (thus <tt>dir</tt>
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  public boolean getDirection ( MicroFacet mf, int order, int total, double[] dir )
  {
    return getDirection( mf, dir );
  }

  //==========================================================================
  //  Module registration:

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "LightSourceToIntersectable";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_DATA + "." + C_LIGHT;

}
