package cz.cuni.jagrlib.iface;

import cz.cuni.jagrlib.*;

/**
 * Abstract light model (BRDF function).
 *
 * @version 0.26 $Rev: 476 $ $Date: 2007-06-03 03:11:16 +0200 (Sun, 03 Jun 2007) $ $Author: pepca $
 * @since   0.11
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/LightModel.java">LightModel.java</a>
 * @see     LightSource
 */
public interface LightModel extends Property
{
  /** Common material property: color. */
  String MATERIAL_COLOR = "Color";

  /** Common material property: ambient coefficient. */
  String MATERIAL_KA    = "Ka";

  /** Common material property: diffuse coefficient. */
  String MATERIAL_KD    = "Kd";

  /** Common material property: specular coefficient. */
  String MATERIAL_KS    = "Ks";

  /** Common material property: specular exponent. */
  String MATERIAL_H     = "H";

  /** Common material property: transparency. */
  String MATERIAL_KT    = "Kt";

  /** Common material property: index of refraction. */
  String MATERIAL_N     = "N";

  /** Component mask: Diffuse component. */
  int COMP_DIFFUSE         = 1;

  /** Component mask: Specular reflection component. */
  int COMP_SPEC_REFLECTION = 2;

  /** Component mask: Specular reflection component. */
  int COMP_SPEC_REFRACTION = 4;

  /** Component mask: both specular components (used in ray-tracing). */
  int COMP_SPECULAR        = COMP_SPEC_REFLECTION + COMP_SPEC_REFRACTION;

  /** Component mask: all components (used with light sources). */
  int COMP_ALL             = COMP_DIFFUSE + COMP_SPECULAR;

  /**
   * Create material object for the shading model. All data are initialized to their
   * default values.
   *
   * @return Data object representing the requested material.
   */
  Property createMaterial ();

  /**
   * Returns the requested BRDF sample (BRDF is treated as monochrome).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  data Non-mandatory data object to hold BRDF state information.
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBRDF(MicroFacet,double[],double[],int)
   * @see    #colorBRDF(double[],double[],double[],Property,int,double[])
   */
  double monoBRDF ( double[] in, double[] out, double[] normal, Property data, int mask );

  /**
   * Returns the requested BRDF sample (BRDF is treated as monochrome). Computations are done
   * relatively to the given {@link MicroFacet} (point on surface of some solid).
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient <b>f(in-&gt;mf-&gt;out)</b>.
   * @see    #monoBRDF(double[],double[],double[],Property,int)
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   * @see    MicroFacet#material
   */
  double monoBRDF ( MicroFacet mf, double[] in, double[] out, int mask );

  /**
   * Returns the requested BRDF sample (BRDF is treated as band-color).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  data Non-mandatory data object to hold BRDF state information.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   * @see    #monoBRDF(double[],double[],double[],Property,int)
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   */
  double[] colorBRDF ( double[] in, double[] out, double[] normal, Property data, int mask, double[] result );

  /**
   * Returns the requested BRDF sample (BRDF is treated as band-color). Computations are done
   * relatively to the given {@link MicroFacet} (point on surface of some solid).
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   * @see    #monoBRDF(MicroFacet,double[],double[],int)
   * @see    #colorBRDF(double[],double[],double[],Property,int,double[])
   * @see    MicroFacet#material
   */
  double[] colorBRDF ( MicroFacet mf, double[] in, double[] out, int mask, double[] result );

  /**
   * Accumulates light contributions from set of light sources through the given
   * {@link MicroFacet} (point on surface of some solid) to the viewing direction.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  lights Set of light sources that can light the point.
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}, otherwise the
   *                initial value will be preseved.
   * @return Output color coefficients.
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   * @see    MicroFacet#material
   */
  double[] lightSum ( MicroFacet mf, int order, int total, LightSource[] lights, double[] out, double[] result );

  /**
   * Retrieves the original color (defining surface color).
   *
   * @param  color Pre-allocated {@code double[C]} array to hold the color
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/). Can
   *               be {@code null}.
   * @param  data Data object with BRDF state information (optional).
   * @return Array containing copy of original color (can be altered).
   */
  double[] getOrigColor ( double[] color, Property data );

  /**
   * Sets the actual color (for the given BRDF data).
   * This color will be used in subsequent BRDF calculations.
   * The color array won't be altered.
   *
   * @param  color Color definition - {@code double[C]} array
   *                ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *                color system/ or greater /for more complex spectral color sampling/).
   * @param  data Data object with BRDF state information.
   */
  void setColor ( double[] color, Property data );

}
