package cz.cuni.jagrlib.piece;

import java.util.Arrays;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Cluster-dot halftoning able to use external "level-set" function.
 * Scales input raster image, gamma-corrects it, converts it to {@link RasterGraphics#MODE_GRAY} and
 * uses only {@code 0.0} and {@code 1.0} values.
 * <p>External <b>level-set function</b> must be 2D real function mapping [-1,1]<sup>2</sup> square
 * to "level-set" distance ({code 0.0} for first painted pixel, only positive values are allowed).
 * Value transfer function and geometric transforms (angle, LPI) will be performed automatically.
 *
 * @version 0.26 $Rev: 823 $ $Date: 2008-12-07 16:35:05 +0100 (Sun, 07 Dec 2008) $ $Author: pepca $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ClusterDotDither.java">ClusterDotDither.java</a>
 */
public class ClusterDotDither extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Implemented internal dot shapes. */
  protected static enum DotType
  {
    /** Round dot (lacking special treatment of the scale end). */
    ROUND,
    /** Round dot with square transition. */
    ROUND_SQUARE,
    /** Rectangle dot (with or w/o external function). */
    RECTANGLE,
    /** External level-set. */
    LEVEL_SET,
  }

  /** Actual dot shape. */
  protected DotType shape    = DotType.ROUND_SQUARE;

  /** Target image width in pixels ({@code 0} for original size). */
  protected int width        = 0;

  /** Target image height in pixels ({@code 0} for original size). */
  protected int height       = 0;

  /** Scale factor ({@code 0.0} or {@code 1.0} for original size). */
  protected double scale     = 1.0;

  /** Simulated gamma-exponent. */
  protected double gamma     = 2.0;

  /** Screen angle in degrees. */
  protected double angle     = 45.0;

  /** Dots per inch of the output image. */
  protected double dpi       = 600.0;

  /** Lines per inch (distance between adjacent dots) of the screen. */
  protected double lpi       = 75.0;

  //--------------------------------------------------------------------------

  /** Transform matrix (2x2). */
  protected double[] m = new double[ 4 ];

  /** Sampling side. */
  protected static final int SAMPLING_SIDE = 100;

  /** Sampling cell size. */
  protected static final double SAMPLING_CELL = 2.0 / SAMPLING_SIDE;

  /** Sampling table size. */
  protected static final int SAMPLING_SIZE = SAMPLING_SIDE * SAMPLING_SIDE;

  /** Transfer function table size. */
  protected static final int TABLE_SIZE = 8000;

  /** Scaling factor for transfer function table (origin is in {@code 0.0}). */
  protected double tableScale;

  /** Transfer function: maps field /levelset/ value to gray level ({@code 0.0} to {@code 1.0}). */
  protected double[] transferTable = new double[ TABLE_SIZE ];

  //==========================================================================
  //  Support:

  /** Level-set function for round dots (w/o transition). */
  protected static class RoundLevelSet extends DefaultRealFunction
  {
    @Override
    public double f ( double x, double y )
    {
      return( x * x + y * y );
    }
  }

  /** Level-set function for round dots with square transition. */
  protected static class RoundSquareLevelSet extends DefaultRealFunction
  {
    @Override
    public double f ( double x, double y )
    {
      double absx = Math.abs( x );
      double absy = Math.abs( y );
      if ( absx + absy <= 1.0 )
        return( absx * absx + absy * absy );

      absx = 1.0 - absx;
      absy = 1.0 - absy;
      return( 2.0 - absx * absx - absy * absy );
    }
  }

  /** Level-set function for square dots (external shape function is not considered yet). */
  protected static class RectangleLevelSet extends DefaultRealFunction
  {
    @Override
    public double f ( double x, double y )
    {
      return( Math.abs( x ) + Math.abs( y ) );
    }
  }

  /** Accelerator: actual level-set function. */
  protected RealFunction field = new RoundLevelSet();

  /** Shared random-number generator for level-set sampling. */
  protected static final RandomJames rnd = new RandomJames();

  /** Sets level-set function and prepares transfer function table. */
  protected void prepareTransferFunction ( RealFunction fld )
  {
    switch ( shape )
    {
      case ROUND:
        field = new RoundLevelSet();
        break;
      case RECTANGLE:
        field = new RectangleLevelSet();
        break;
      case LEVEL_SET:
        field = fld;
        break;
      default:
        field = new RoundSquareLevelSet();
    }

      // Transfer function:

      // I. random sampling of the [-1,1] square
    double[] s = new double[ SAMPLING_SIZE ];
    int ix, iy, i;
    double ox, oy, x, y, f;
    double minf = Double.MAX_VALUE;
    double maxf = Double.MIN_VALUE;
    for ( iy = i = 0, oy = -1.0; iy++ < SAMPLING_SIDE; oy += SAMPLING_CELL )
      for ( ix = 0, ox = -1.0; ix++ < SAMPLING_SIDE; ox += SAMPLING_CELL )
      {
        x = ox + SAMPLING_CELL * rnd.uniformNumber();
        y = oy + SAMPLING_CELL * rnd.uniformNumber();
        s[ i++ ] = f = field.f( x, y );
        if ( f < minf ) minf = f;
        if ( f > maxf ) maxf = f;
      }

      // II. compute the inverse table
      // 0.0 to maxf values were found.. there will be mapped to the [0.0,1.0] segment
    tableScale = TABLE_SIZE / maxf;
    final double grayScale = 1.0 / SAMPLING_SIZE;
    final double xScale = maxf / TABLE_SIZE;
    Arrays.sort( s );
    ix = 0;
    int j;
    for ( j = 0, x = 0.0; j < TABLE_SIZE; x += xScale, j++ )
    {     // look for the 'x' value in the sampling array, its position should be
          // encoded into the 'transferTable'..
      if ( x <= minf )
        transferTable[ j ] = 0.0;
      else
      if ( x >= maxf )
        transferTable[ j ] = 1.0;
      else
      {
        i = Arrays.binarySearch( s, ix, SAMPLING_SIZE, x );
        if ( i >= 0 )
          transferTable[ j ] = i * grayScale;
        else
        {
          i = -i - 2;
          if ( i >= SAMPLING_SIZE - 1 )
            transferTable[ j ] = 1.0;
          else                              // use linear interpolation between s[i] and s[i+1]
          {
            final double denom = s[i+1] - s[i];
            if ( denom <= Double.MIN_VALUE )
              transferTable[ j ] = (i + 0.5) * grayScale;
            else
              transferTable[ j ] = (i + (x - s[i]) / denom) * grayScale;
          }
        }
        ix = i;
        if ( ix >= SAMPLING_SIZE ) ix = SAMPLING_SIZE - 1;
      }
    }

      // Transform matrix:

      // distance between adjacent dots in pixels: dpi / lpi
      // distance between adjacent dots in level-set space: 2.0
    final double globalScale = 2.0 * lpi / dpi;
    m[0] =   m[3] = globalScale * Math.cos( Math.toRadians( angle ) );
    m[1] = -(m[2] = globalScale * Math.sin( Math.toRadians( angle ) ));
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transform.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1",    IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2",    IFACE+"RasterGraphics" );
                         field  = (RealFunction)  getInterface( PL_FUNCTION, IFACE+"RealFunction" );

    if ( output == null ) return false;

      // prepare transfer function table:
    prepareTransferFunction( field );

      // output image:
    int iWid = (input == null) ? 1000 : input.getWidth();
    int iHei = (input == null) ?  250 : input.getHeight();
    int oWid = width;
    int oHei = height;

    if ( oWid == 0 )
      if ( Geometry.isZero( scale ) ||
           Geometry.isZero( scale - 1.0 ) )
        oWid = iWid;
      else
        oWid = Formula.round( iWid * scale );
    if ( oWid < 2 ) oWid = 2;

    if ( oHei == 0 )
      if ( Geometry.isZero( scale ) ||
           Geometry.isZero( scale - 1.0 ) )
        oHei = iHei;
      else
        oHei = Formula.round( iHei * scale );
    if ( oHei < 2 ) oHei = 2;

    output.init( oWid, oHei, RasterGraphics.MODE_GRAY, 0 );

    int ox, oy;                           // output coordinates

      // support variables for input image reading:
    final double dGray = 1.0 / (oWid - 1.0);
    final double kx = (iWid - 1.0) / (oWid - 1.0); // src_x = kx * target_x
    final double ky = (iHei - 1.0) / (oHei - 1.0); // src_y = ky * target_y
    double x, y;                          // input coordinates
    double pix00, pix10, pix01, pix11;
    int    iy, ix;
    double fx, fy, fx1, fy1;

      // error distribution:
    double pixel;

    for ( oy = 0; oy < oHei; oy++ )
      for ( ox = 0; ox < oWid; ox++ )
      {
        if ( input == null )              // void input => horizontal vignette
          pixel = 1.0 - ox * dGray;

        else
        if ( iWid == oWid &&
             iHei == oHei )               // dithering (no scaling)
          pixel = Math.pow( input.getGrayDouble( ox, oy ), gamma );

        else                              // halftoning (scaling)
        {
          x   = ox * kx;                  // input-image coordinate
          fx  = Math.floor( x );
          ix  = (int)fx;                  // x = ix + fx
          fx  = x - fx;
          fx1 = 1.0 - fx;

          y   = oy * ky;                  // input-image coordinate
          fy  = Math.floor( y );
          iy  = (int)fy;                  // y = iy + fy
          fy  = y - fy;
          fy1 = 1.0 - fy;

            // four surrounding pixels:
          pix00 = Math.pow( input.getGrayDouble( ix,     iy     ), gamma );
          pix10 = Math.pow( input.getGrayDouble( ix + 1, iy     ), gamma );
          pix01 = Math.pow( input.getGrayDouble( ix,     iy + 1 ), gamma );
          pix11 = Math.pow( input.getGrayDouble( ix + 1, iy + 1 ), gamma );

            // bilinear interpolation:
          pixel = fx1 * ( fy1 * pix00 + fy * pix01 ) +
                  fx  * ( fy1 * pix10 + fy * pix11 );
        }

          // transfer function lookup:
        x = Formula.remainder( ox * m[0] + oy * m[1] + 1.0, 2.0 ) - 1.0;
        y = Formula.remainder( ox * m[2] + oy * m[3] + 1.0, 2.0 ) - 1.0;
        ix = Formula.round( tableScale * field.f( x, y ) );
        if ( ix >= TABLE_SIZE ) ix = TABLE_SIZE - 1;

        output.putPixel( ox, oy, transferTable[ ix ] < (1.0 - pixel) ? 0.0 : 1.0 );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String SHAPE     = "Dot shape";
  public final static String ANGLE     = "Angle";
  public final static String DPI       = "DPI";
  public final static String LPI       = "LPI";
  public final static String WIDTH     = "Width";
  public final static String HEIGHT    = "Height";
  public final static String SCALE     = "Scale";
  public final static String GAMMA     = "Gamma";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SHAPE:
    if ( key.compareTo(SHAPE) == 0 )
    {
      shape = enumProperty( value, shape );
      return;
    }
      // ANGLE:
    if ( key.compareTo(ANGLE) == 0 )
    {
      angle = doubleProperty( value, angle, 0.01, 100.0 );
      return;
    }
      // DPI:
    if ( key.compareTo(DPI) == 0 )
    {
      dpi = doubleProperty( value, dpi, 1.0, 1.e6 );
      return;
    }
      // LPI:
    if ( key.compareTo(LPI) == 0 )
    {
      lpi = doubleProperty( value, lpi, 1.0, 1.e6 );
      return;
    }
      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      width = intProperty( value, width, 0, 1000000 );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      height = intProperty( value, height, 0, 1000000 );
      return;
    }
      // SCALE:
    if ( key.compareTo(SCALE) == 0 )
    {
      scale = doubleProperty( value, scale, 0.0, 1000.0 );
      return;
    }
      // GAMMA:
    if ( key.compareTo(GAMMA) == 0 )
    {
      gamma = doubleProperty( value, gamma, 0.01, 100.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SHAPE) == 0 )
      return shape.ordinal();
    if ( key.compareTo(ANGLE) == 0 )
      return angle;
    if ( key.compareTo(DPI) == 0 )
      return dpi;
    if ( key.compareTo(LPI) == 0 )
      return lpi;
    if ( key.compareTo(WIDTH) == 0 )
      return width;
    if ( key.compareTo(HEIGHT) == 0 )
      return height;
    if ( key.compareTo(SCALE) == 0 )
      return scale;
    if ( key.compareTo(GAMMA) == 0 )
      return gamma;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Cluster-dot halftoning";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRealFunctionAndRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Cluster-dot halftoning capable of several dot shapes.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,    IFACE+"Trigger" );
      t.newOptOutputPlug( "image1",    IFACE+"RasterGraphics" );
      t.newOutputPlug(    "image2",    IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_FUNCTION, IFACE+"RealFunction" );
        // Property:
      t.propBegin( SHAPE, TYPE_INTEGER, "Dot shape", true );
        t.propDefault( DotType.ROUND_SQUARE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Round",                     DotType.ROUND.ordinal(),        "Round with bad scale end" );
        t.propEnum( "Round (square transition)", DotType.ROUND_SQUARE.ordinal(), "Round with square transition in the middle" );
        t.propEnum( "Rectangle",                 DotType.RECTANGLE.ordinal(),    "Rectangle (optional external function)" );
        t.propEnum( "Level set",                 DotType.LEVEL_SET.ordinal(),    "External level set function" );
      t.propEnd();
      t.propBegin( ANGLE, TYPE_DOUBLE, "Screen angle in degrees", true );
        t.propDefault( 45.0 );
        t.propBounds(  0.0, 360.0 );
      t.propEnd();
      t.propBegin( DPI, TYPE_DOUBLE, "DPI of output image", true );
        t.propDefault( 600.0 );
        t.propBounds(  1.0, 1.e6 );
      t.propEnd();
      t.propBegin( LPI, TYPE_DOUBLE, "LPI of the screen (distance between adjacent dots)", true );
        t.propDefault( 75.0 );
        t.propBounds(  1.0, 1.e6 );
      t.propEnd();
      t.propBegin( WIDTH, TYPE_INTEGER, "Output width", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Output height", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( SCALE, TYPE_DOUBLE, "Image scale factor", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1000.0 );
      t.propEnd();
      t.propBegin( GAMMA, TYPE_DOUBLE, "Simulated gamma-exponent", true );
        t.propDefault( 2.0 );
        t.propBounds(  0.01, 100.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
