package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * CSG representation of 3D scenes for Ray-tracing rendering methods.
 * Dynamic scene (time-dependency via {@link TimeDependent} interface).
 *
 * @version 0.26 $Rev: 919 $ $Date: 2010-04-08 03:45:53 +0200 (Thu, 08 Apr 2010) $ $Author: pepca $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/DynamicCSGScene.java">DynamicCSGScene.java</a>
 */
public class DynamicCSGScene extends StaticCSGScene
{
  //==========================================================================
  //  Data:

  /** Current virtual time in seconds. */
  protected double time = 0.0;

  /** Upper bound for time. */
  protected double maxTime = 100.0;

  //==========================================================================
  //  Support:

  /** All scene components have be usable. */
  @Override
  protected void assertScene ()
  {
    super.assertScene();
    setTime( time );
  }

  //==========================================================================
  //  Interface Intersectable:

  //  Inherited completely

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  @Override
  public double[] getTimeInterval ( double[] pre )
  {
    if ( pre == null ||
         pre.length < 2 )
      pre = new double[ 2 ];

    pre[0] = 0.0;
    pre[1] = maxTime;
    return pre;
  }

  /**
   * Returns the actual time value for the scene (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  @Override
  public double getTime ()
  {
    return time;
  }

  /**
   * Sets the actual time value for the scene (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  @Override
  public void setTime ( double time )
  {
    this.time = time;

    if ( root instanceof TimeDependent )
      ((TimeDependent)root).setTime( time );

    if ( camera instanceof TimeDependent )
      ((TimeDependent)camera).setTime( time );

    if ( lights != null )
      for ( LightSource ls : lights )
        if ( ls instanceof TimeDependent )
          ((TimeDependent)ls).setTime( time );
  }

  //==========================================================================
  //  Interface RTScene:

  //  Inherited completely

  //==========================================================================
  //  Interface Property:

  public final static String TIME     = "Time";
  public final static String MAX_TIME = "MaxTime";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // TIME:
    if ( key.compareTo(TIME) == 0 )
    {
      time = doubleProperty( value, time, 0.0, maxTime );
      setTime( time );
      return;
    }

      // MAX_TIME:
    if ( key.compareTo(MAX_TIME) == 0 )
    {
      maxTime = doubleProperty( value, maxTime );
      if ( maxTime < 0.0 ) maxTime = 1.0;
      return;
    }

    super.set( key, value );
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(TIME) == 0 )
      return time;

    if ( key.compareTo(MAX_TIME) == 0 )
      return maxTime;

    return super.get( key );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "CSG-scene (dynamic)";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RTScene";

  /** Object description. */
  private final static String DESCRIPTION = "CSG representation of 3D scene for Ray-tracing " +
                                            "(dynamic implementation).";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_DATA + '.' + C_SCENE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,        IFACE+"RTScene" );
      t.newInputPlug(     PL_INTERSECTION, IFACE+"Intersectable" );
      t.newInputPlug(     PL_PROPERTY,     IFACE+"Property" );
      t.newOptOutputPlug( PL_OUTPUT,       IFACE+"RayGenerator" );
      t.newOptOutputPlug( PL_DATA,         IFACE+"Brep" );
      t.newOptOutputPlug( PL_IMPORT,       IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( SCENE, TYPE_OBJECT, "Scene definition file-name", true );
        t.propDefault( "data/scene1.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
      t.propBegin( TIME, TYPE_DOUBLE, "Current time", true );
        t.propDefault( 0.0 );
      t.propEnd();
      t.propBegin( MAX_TIME, TYPE_DOUBLE, "Maximum time", true );
        t.propDefault( 100.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
