package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Phong shading model (with constant or Schlick K() term).
 *
 * @version 0.26 $Rev: 477 $ $Date: 2007-06-03 03:19:26 +0200 (Sun, 03 Jun 2007) $ $Author: pepca $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/PhongModel.java">PhongModel.java</a>
 */
public class PhongModel extends Piece implements LightModel
{
  //==========================================================================
  //  Support:

  /**
   * Common computation for monochrome BRDF sample (BRDF is treated as monochrome).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  d BRDF parameters.
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient {@code f(in->out)}.
   */
  protected double mono ( double[] in, double[] out,
                          double[] normal, PhongMaterial d, int mask )
  {
    final boolean viewOut = out[0] * normal[0] + out[1] * normal[1] + out[2] * normal[2] > 0.0;

    if ( in == null )                     // ambient term only..
        // dim ambient light if viewer is inside
      return( d.color[0] * (viewOut ? d.ka : (d.ka * d.kt)) );

      // directional light source:
    final double  cosAlpha = in[0] * normal[0] + in[1] * normal[1] + in[2] * normal[2];
    final boolean lightOut = cosAlpha > 0.0;
    double ks = d.ks;
    double kd = d.kd;
    double kt = d.kt;

      // Schlick's adjustments:
    if ( d.sch > Geometry.EPSILON )
    {
      final double oneU    = 1.0 - Math.abs( cosAlpha );     //  1-u
      final double oneUsq  = oneU * oneU;                    // (1-u)^2
      final double schlick = d.sch * oneUsq * oneUsq * oneU; // (1-u)^5 * d.sch
      final double schlickComp = 1.0 - schlick;
      ks += schlick * (kd + kt);
      kd *= schlickComp;
      kt *= schlickComp;
    }

    double[] r = null;
    double coef = 1.0;
    if ( viewOut == lightOut )          // viewer and source are on the same side..
    {
      if ( (mask & COMP_SPEC_REFLECTION) != 0 )
      {
        final double cos2 = cosAlpha + cosAlpha;
        r = new double[]                // reflected vector
        {
          cos2 * normal[0] - in[0],
          cos2 * normal[1] - in[1],
          cos2 * normal[2] - in[2]
        };

        if ( !lightOut &&               // total reflection check
             -cosAlpha <= d.cosTotal )
          if ( (ks += kt) + kd > 1.0 )
            ks = 1.0 - kd;
      }
    }
    else                                // opposite sides => use specular refraction
    {
      if ( (mask & COMP_SPEC_REFRACTION) != 0 )
        r = Geometry.specularRefraction( normal, d.n, in, null );
      coef = kt;
    }

    final double diffuse = (mask & COMP_DIFFUSE) != 0 ?
                           coef * kd * Math.abs( cosAlpha ) : 0.0;
    double specular = 0.0;
    if ( r != null )
    {
      final double cosBeta = r[0] * out[0] + r[1] * out[1] + r[2] * out[2];
      if ( cosBeta > 0.0 )
        specular = coef * ks * Formula.power( cosBeta, d.h );
    }

    return( diffuse * d.color[0] + specular );
  }

  /**
   * Common computation for color BRDF sample (BRDF is treated as band-color).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  d BRDF parameters.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                {@code f(in->out)} ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   */
  protected double[] color ( double[] in, double[] out,
                             double[] normal, PhongMaterial d, int mask,
                             double[] result )
  {
    final int     bands   = d.color.length;
    final boolean viewOut = out[0] * normal[0] + out[1] * normal[1] + out[2] * normal[2] > 0.0;
    int i;
    double coef;

    if ( in == null )                     // ambient term only..
    {
      if ( result == null ||
           result.length < bands )
        result = new double[ bands ];

        // dim ambient light if viewer is inside
      coef = viewOut ? d.ka : (d.ka * d.kt);
      for ( i = 0; i < bands; i++ )
        result[i] = coef * d.color[i];

      return result;
    }

      // directional light source:
    final double  cosAlpha = in[0] * normal[0] + in[1] * normal[1] + in[2] * normal[2];
    final boolean lightOut = cosAlpha > 0.0;
    double ks = d.ks;
    double kd = d.kd;
    double kt = d.kt;

      // Schlick's adjustments:
    if ( d.sch > Geometry.EPSILON )
    {
      final double oneU    = 1.0 - Math.abs( cosAlpha );     //  1-u
      final double oneUsq  = oneU * oneU;                    // (1-u)^2
      final double schlick = d.sch * oneUsq * oneUsq * oneU; // (1-u)^5 * d.sch
      final double schlickComp = 1.0 - schlick;
      ks += schlick * (kd + kt);
      kd *= schlickComp;
      kt *= schlickComp;
    }

    double[] r = null;
    coef       = 1.0;
    if ( viewOut == lightOut )            // viewer and source are on the same side..
    {
      if ( (mask & COMP_SPEC_REFLECTION) != 0 )
      {
        final double cos2 = cosAlpha + cosAlpha;
        r = new double[]                    // reflected vector
          {
            cos2 * normal[0] - in[0],
            cos2 * normal[1] - in[1],
            cos2 * normal[2] - in[2]
          };

        if ( !lightOut &&                   // total reflection check
             -cosAlpha <= d.cosTotal )
          if ( (ks += kt) + kd > 1.0 )
            ks = 1.0 - kd;
      }
    }
    else                                  // opposite sides => use specular refraction
    {
      if ( (mask & COMP_SPEC_REFRACTION) != 0 )
        r = Geometry.specularRefraction( normal, d.n, in, null );
      coef = kt;
    }

    if ( result == null ||
         result.length < bands )
      result = new double[ bands ];

    final double diffuse = (mask & COMP_DIFFUSE) != 0 ?
                           coef * kd * Math.abs( cosAlpha ) : 0.0;
    double specular = 0.0;
    if ( r != null )
    {
      final double cosBeta = r[0] * out[0] + r[1] * out[1] + r[2] * out[2];
      if ( cosBeta > 0.0 )
        specular = coef * ks * Formula.power( cosBeta, d.h );
    }

    for ( i = 0; i < bands; i++ )
      result[i] = diffuse * d.color[i] + specular;

    return result;
  }

  //==========================================================================
  //  Interface LightModel:

  /**
   * Create material object for the Phong shading model.
   * All data are initialized to their default values.
   *
   * @return Data object representing the requested material.
   */
  public Property createMaterial ()
  {
    return new PhongMaterial();
  }

  /**
   * Returns the requested BRDF sample (BRDF is treated as monochrome).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  data Non-mandatory data object to hold BRDF state information.
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient {@code f(in->out)}.
   * @see    #monoBRDF(MicroFacet,double[],double[],int)
   * @see    #colorBRDF(double[],double[],double[],Property,int,double[])
   */
  public double monoBRDF ( double[] in, double[] out, double[] normal, Property data, int mask )
  {
    if ( data == null ||
         !(data instanceof PhongMaterial) ) return 0.0;

    return mono( in, out, normal, (PhongMaterial)data, mask );
  }

  /**
   * Returns the requested BRDF sample (BRDF is treated as monochrome). Computations are done
   * relatively to the given {@code MicroFacet} (point on surface of some solid).
   *
   * @param  mf The {@code MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient <b>f(in-&gt;mf-&gt;out)</b>.
   * @see    #monoBRDF(double[],double[],double[],Property,int)
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   * @see    MicroFacet#material
   */
  public double monoBRDF ( MicroFacet mf, double[] in, double[] out, int mask )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL |
                              MicroFacet.VALID_MATERIAL ) != 0 ||
         !(mf.material instanceof PhongMaterial) ) return 0.0;

    return mono( in, out, mf.normal, (PhongMaterial)(mf.material), mask );
  }

  /**
   * Returns the requested BRDF sample (BRDF is treated as band-color).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  data Non-mandatory data object to hold BRDF state information.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                {@code f(in->out)} ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   * @see    #monoBRDF(double[],double[],double[],Property,int)
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   */
  public double[] colorBRDF ( double[] in, double[] out, double[] normal, Property data, int mask,
                              double[] result )
  {
    if ( data == null ||
         !(data instanceof PhongMaterial) ) return null;

    return color( in, out, normal, (PhongMaterial)data, mask, result );
  }

  /**
   * Returns the requested BRDF sample (BRDF is treated as band-color). Computations are done
   * relatively to the given {@code MicroFacet} (point on surface of some solid).
   *
   * @param  mf The {@code MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   * @see    #monoBRDF(MicroFacet,double[],double[],int)
   * @see    #colorBRDF(double[],double[],double[],Property,int,double[])
   * @see    MicroFacet#material
   */
  public double[] colorBRDF ( MicroFacet mf, double[] in, double[] out, int mask,
                              double[] result )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL   |
                              MicroFacet.VALID_MATERIAL |
                              MicroFacet.VALID_COLOR ) != 0 ||
         !(mf.material instanceof PhongMaterial) ) return null;

    setColor( mf.color, (PhongMaterial)(mf.material) );
    return color( in, out, mf.normal, (PhongMaterial)(mf.material), mask, result );
  }

  /**
   * Accumulates light contributions from set of light sources through the given
   * {@code MicroFacet} (point on surface of some solid) to the viewing direction.
   *
   * @param  mf The {@code MicroFacet} (point on surface of some solid).
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  lights Set of light sources that can light the point.
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}, otherwise the
   *                initial value will be preseved.
   * @return Output color coefficients.
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   * @see    MicroFacet#material
   */
  public double[] lightSum ( MicroFacet mf, int order, int total,
                             LightSource[] lights, double[] out,
                             double[] result )
  {
    if ( lights == null ||
         mf     == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL   |
                              MicroFacet.VALID_MATERIAL |
                              MicroFacet.VALID_COLOR ) != 0 ||
         !(mf.material instanceof PhongMaterial) ) return null;

    PhongMaterial d = (PhongMaterial)(mf.material);
    setColor( mf.color, d );
    int bands = d.color.length;

      // initialize the results[] array:
    if ( result == null || result.length < bands )
      result = new double[ bands ];

      // temporary data:
    double[] lDir  = new double[ 4 ];     // light source direction vector
    double[] lInt  = new double[ bands ]; // light source intensity
    double[] contr = new double[ bands ]; // contribution of an individual ray

      // accumulate light from light sources:
    int i;
    for ( LightSource light : lights )
    {
      if ( light == null ) break;

      if ( !light.getIntensity( mf, order, total, lInt ) )
        continue;

      boolean dirSource = light.getDirection( mf, order, total, lDir );
      if ( color( dirSource ? lDir : null, out, mf.normal, d, COMP_ALL, contr ) == null )
        continue;

      for ( i = 0; i < bands; i++ )
        result[i] += contr[i] * lInt[i];
    }

    return result;
  }

  /**
   * Retrieves the original color (defining surface color).
   *
   * @param  color Pre-allocated {@code double[C]} array to hold the color
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/). Can
   *               be {@code null}.
   * @param  data Data object with BRDF state information (optional).
   * @return Array containing copy of original color (can be altered).
   */
  public double[] getOrigColor ( double[] color, Property data )
  {
    if ( data == null ||
         !(data instanceof PhongMaterial) ) return null;

    PhongMaterial d = (PhongMaterial)data;
    if ( d.origColor == null ) return null;

    int bands = d.origColor.length;
    if ( color == null ||
         color.length < bands )
      color = Formula.cloneArray( d.origColor );
    else
      System.arraycopy( d.origColor, 0, color, 0, bands );

    return color;
  }

  /**
   * Sets the actual color (for the given BRDF data).
   * This color will be used in subsequent BRDF calculations.
   * The color array won't be altered.
   *
   * @param  color Color definition - {@code double[C]} array
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/).
   * @param  data Data object with BRDF state information.
   */
  public void setColor ( double[] color, Property data )
  {
    if ( data != null &&
         data instanceof PhongMaterial )
      ((PhongMaterial)data).color = color;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Phong shading model";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "LightModel";

  /** Object description. */
  private final static String DESCRIPTION = "Phong shading model with constant K() term.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_LIGHT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"LightModel" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
