package cz.cuni.jagrlib.piece;

import java.util.*;
import java.awt.Point;
import java.awt.geom.Point2D;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Scan-line rasterization of 2D polygon. Implements both simple and anti-aliased rendering.
 *
 * @version 0.26 $Rev: 813 $ $Date: 2008-10-21 15:44:11 +0200 (Tue, 21 Oct 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/PolygonScanline.java">PolygonScanline.java</a>
 */
public class PolygonScanline extends Piece implements PolygonFillRender
{
  //==========================================================================
  //  Support classes:

  /**
   * Polygon edge.
   */
  protected static class PolyEdge implements Comparable< PolyEdge >
  {
    /** Starting Y-coordinate. */
    public int y0;

    /** X-coordinate of the edge-hline intersection. */
    public double x;

    /** X difference between two adjacent hlines. */
    public double dx;

    /** Number of hlines to draw. */
    public int count;

    /** Edge orientation. */
    public boolean ori;

    public PolyEdge ( double x1, double y1, double x2, double y2 )
    {
      double yy0;
      if ( y1 < y2 )                      // top-down edge
      {
        yy0   = Math.ceil( y1 );
        y0    = Formula.round( yy0 );
        dx    = (x2 - x1) / (y2 - y1);
        x     = x1 + (yy0 - y1) * dx;
        count = Formula.round( Math.ceil( y2 ) - yy0 );
        ori   = true;
      }
      else                                // bottom-up edge
      {
        yy0   = Math.ceil( y2 );
        y0    = Formula.round( yy0 );
        dx    = (x1 - x2) / (y1 - y2);
        x     = x2 + (yy0 - y2) * dx;
        count = Formula.round( Math.ceil( y1 ) - yy0 );
        ori   = false;
      }
    }

    /** Natural ordering: [ y0, x ]. */
    public int compareTo ( PolyEdge o )
    {
      if ( o == null )
        throw new NullPointerException();

      if ( this == o )
        return 0;

      if ( y0 != o.y0 )
        return( y0 < o.y0 ? -1 : 1 );

      if ( x != o.x )
        return( x < o.x ? -1 : 1 );

      return( hashCode() < o.hashCode() ? -1 : 1 );
    }
  }

  /** Edge comparator for actual edge list: [ x ]. */
  protected static class EdgeComparator implements Comparator< PolyEdge >
  {
    public int compare ( PolyEdge a, PolyEdge b )
    {
      if ( a == b )
        return 0;

      if ( a.x != b.x )
        return( a.x < b.x ? -1 : 1 );

      return( a.hashCode() < b.hashCode() ? -1 : 1 );
    }
  }

  /** Single comparator instance. */
  protected final static EdgeComparator eComparator = new EdgeComparator();

  /**
   * Temporary rendering target for super-sampling.
   * Flushes finished hlines to attached {@link AlphaMask}.
   * <p>Uses PE's linear super-sampling buffer.
   */
  protected static class Supersampler implements BitMaskCore
  {
    /** Buffer granularity (to reduce memory allocations). */
    protected final static int BUFFER_GRANUL = 16;

    /** Supersampling factor (accelerator). */
    protected final int supersampling;

    /** Pixel area multiplier (accelerator). */
    protected final double areaCoef;

    /** Anti-aliased output. */
    protected final AlphaMask output;

    /** Linear ss-buffer. */
    protected int[] buffer;

    /** Left buffer coordinate ({@link #supersampling}-multiple). */
    protected int xMin;

    /** Right buffer coordinate. */
    protected int xMax;

    /** Upper buffer coordinate ({@link #supersampling}-multiple). */
    protected int yMin;

    public Supersampler ( int superSampling, AlphaMask aMask )
    {
      supersampling = superSampling;
      areaCoef      = 1.0 / (supersampling * supersampling);
      output        = aMask;
      buffer        = null;
      yMin          =
      xMin          =
      xMax          = 0;
    }

    //==========================================================================
    //  Support:

    /** Flushes unwritten data to attached {@link AlphaMask}. */
    public void flushBuffer ()
    {
      if ( buffer == null ||
           xMax <= xMin ) return;

      final int x0 = xMin / supersampling;
      final int y0 = yMin / supersampling;
      for ( int i = 0, x = xMin; x < xMax; x += supersampling, i++ )
        if ( buffer[i] > 0 )
          output.putPixel( x0 + i, y0, buffer[i] * areaCoef );

      init();
    }

    //==========================================================================
    //  Interface BitMaskCore:

    /**
     * Bit-mask (re-)initialization.
     *
     * @param  x Width of a new bitmask.
     * @param  y Height of a new bitmask.
     */
    public void init ( int x, int y )
    {
      init();
    }

    /**
     * Bit-mask re-initialization (keeps current bitmask size - if the size
     * is relevant).
     */
    public void init ()
    {
      xMin = xMax = 0;
      if ( buffer != null )
        Arrays.fill( buffer, 0 );
    }

    /**
     * Sets the given pixel's value to {@code true} = "painted".
     * Accepts current <b>binary set operation</b>.
     *
     * @param  x X coordinate of a pixel.
     * @param  y Y coordinate of a pixel.
     */
    public void setPixel ( int x, int y )
    {
      setHLine( x, x + 1, y );
    }

    /**
     * Returns the pixel's value.
     *
     * @param  x X coordinate of a pixel.
     * @param  y Y coordinate of a pixel.
     * @return Pixel value ({@code false} for off-canvas access).
     */
    public boolean getPixel ( int x, int y )
    {
      return false;
    }

    /**
     * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
     * Accepts current <b>binary set operation</b>.
     *
     * @param  x1 X coordinate of the starting pixel.
     * @param  x2 X coordinate of the first pixel after the line.
     * @param  y Common Y coordinate of the line.
     */
    public void setHLine ( int x1, int x2, int y )
    {
      if ( y < yMin || y >= yMin + supersampling )
      {                                   // out of the current line
        if ( xMax > xMin )                // anything to flush..
          flushBuffer();

        yMin = y - y % supersampling;
      }

      int xFloor = x1 - x1 % supersampling;

        // OK, I'm on the right line..
      if ( x1 < xMin ||
           x2 > xMax )                    // buffer needs to be extended
      {
        int len = Formula.granularity( (x2 + supersampling - 1 - xFloor) / supersampling, BUFFER_GRANUL );
        if ( xMin >= xMax )               // previous buffer was clean
        {
          xMin = xFloor;
          xMax = x2;
          if ( buffer == null ||          // new buffer array
               buffer.length < len )
            buffer = new int[ len ];
          else
            Arrays.fill( buffer, 0 );
        }
        else                              // there was something in the buffer before
        {
          int xMinimum = Math.min( xMin, xFloor );
          int xMaximum = Math.max( xMax, x2 );
          int iTo      = (xMin - xMinimum) / supersampling;
          len = Formula.granularity( (xMaximum + supersampling - 1 - xMinimum) / supersampling, BUFFER_GRANUL );

          if ( len > buffer.length )      // new buffer array
          {
            int[] newBuffer = new int[ len ];
            System.arraycopy( buffer, 0, newBuffer, iTo, Math.min( buffer.length, len - iTo ) );
            buffer = newBuffer;
          }
          else                            // current buffer is long enough
            if ( iTo > 0 )
            {
              System.arraycopy( buffer, 0, buffer, iTo, buffer.length - iTo );
              Arrays.fill( buffer, 0, iTo, 0 );
            }

          xMin = xMinimum;
          xMax = xMaximum;
        }
      }

        // write data into the buffer:
      int i = (xFloor - xMin) / supersampling;
      do
      {
        int contrib = (xFloor > x1) ? supersampling : xFloor + supersampling - x1;
        if ( (xFloor += supersampling) > x2 )
          contrib -= xFloor - x2;
        buffer[ i++ ] += contrib;
      }
      while ( xFloor < x2 );
    }

    /** Useless enumerator singleton. */
    protected final static BitMaskEnumerator ENUM = new EmptyBitMaskEnumerator();

    /**
     * Bit-mask enumeration.
     *
     * @return The {@link BitMaskEnumerator} object instance.
     */
    public BitMaskEnumerator enumerator ()
    {
      return ENUM;
    }
  }

  //==========================================================================
  //  Data:

  /**
   * Actual inside-test mode.
   *
   * @see    #POLY_INSIDE
   * @see    #POLY_INSIDE_PARITY
   * @see    #POLY_INSIDE_WINDING
   */
  protected int insideRule = POLY_INSIDE_PARITY;

  /** Supersampling factor. */
  protected int supersampling = 4;

  //==========================================================================
  //  Interface PolygonFillRender:

  /**
   * Fills the given polygon.
   * Uses pre-defined fill style.
   *
   * @param  v Array of polygon vertices (needs not to be closed, {@code null} can interrupt the sequence).
   * @see    #POLY_INSIDE
   */
  public void fillPolygon ( Point[] v )
  {
    if ( v        == null ||
         v.length < 3 ) return;

    Point2D.Double[] d = new Point2D.Double[ v.length ];
    int i;
    for ( i = 0; i < v.length && v[i] != null; i++ )
      d[i] = new Point2D.Double( v[i].x, v[i].y );

    fillPolygon( d );
  }

  /**
   * Fills the given polygon.
   * Uses pre-defined fill style.
   *
   * @param  v Array of polygon vertices (needs not to be closed, {@code null} can interrupt the sequence).
   * @see    #POLY_INSIDE
   */
  public void fillPolygon ( Point2D.Double[] v )
  {
    final BitMask   bm = (BitMask)  getInterface( PL_BITMASK,   IFACE+"BitMask" );
    final AlphaMask am = (AlphaMask)getInterface( PL_ALPHAMASK, IFACE+"AlphaMask" );

    if ( bm == null && am == null ||
         v  == null ||
         v.length < 3 ) return;

      // 1. output object
    Supersampler ss = null;
    BitMaskCore output;
    if ( bm != null )
      output = bm;
    else
      output = ss = new Supersampler( supersampling, am );

      // input edge set sorted by [ y0 ]
    TreeSet< PolyEdge > input = new TreeSet< PolyEdge >();

      // 2. create sorted set of edge records .. input
    int i, i1;
    for ( i = 0; i < v.length && v[i] != null; i++ )
    {                                     // edge from v[i] to v[i+1]
      if ( (i1 = i + 1) >= v.length ||
           v[i1] == null )
        i1 = 0;

      double x1 = v[i].x;
      double y1 = v[i].y;
      double x2 = v[i1].x;
      double y2 = v[i1].y;

      if ( x1 == x2 && y1 == y2 )
        continue;

      if ( ss != null )
      {
        x1 = (x1 + 0.5) * supersampling - 0.5;
        y1 = (y1 + 0.5) * supersampling - 0.5;
        x2 = (x2 + 0.5) * supersampling - 0.5;
        y2 = (y2 + 0.5) * supersampling - 0.5;
      }

      if ( Math.ceil( y1 ) == Math.ceil( y2 ) )
        continue;                         // discarding horizontal edges

      input.add( new PolyEdge( x1, y1, x2, y2 ) );
    }

      // 3. initialize actual edge set
    if ( input.size() == 0 ) return;
    PolyEdge act = input.first();
    input.remove( act );                  // !!! TODO(1.6): act = input.pollFirst()

    TreeSet< PolyEdge > actual = new TreeSet< PolyEdge >( eComparator );
    TreeSet< PolyEdge > backup = new TreeSet< PolyEdge >( eComparator );
    int y = act.y0;
    actual.add( act );

      // 4. hline loop
    do
    {
        // move all relevant edges from input into actual:
      while ( input.size() > 0 &&
              (act = input.first()).y0 == y )
      {                                   // !!! TODO(1.6): act = input.pollFirst()
        input.remove( act );
        actual.add( act );
      }

      int oldX = 0;
      int actX;
      int rule = 0;
      do
      {
          // 4a. edge contribution to the current hline
        act  = actual.first();
        actual.remove( act );             // !!! TODO(1.6): act = actual.pollFirst()
        actX = Formula.round( Math.ceil( act.x ) );

        if ( rule != 0 &&
             actX > oldX )                // draw hline segment from the previous edge..
          output.setHLine( oldX, actX, y );

        if ( insideRule == POLY_INSIDE_PARITY )
          rule  = 1 - rule;
        else
          rule += (act.ori ? 1 : -1);
        oldX = actX;

          // 4b. edge ages for the next hline
        if ( --act.count > 0 )
        {
          act.x += act.dx;
          backup.add( act );
        }
      }
      while ( actual.size() > 0 );

        // swap the actual sets:
      TreeSet< PolyEdge > tmp = backup;
      backup = actual;
      actual = tmp;

        // Y-coordinate of the next line
      y++;
    }
    while ( actual.size() > 0 ||
            input.size()  > 0 );

    if ( ss != null )
      ss.flushBuffer();
  }

  //==========================================================================
  //  Interface Property:

  public final static String SUPERSAMPLING  = "Supersampling factor";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // POLY_INSIDE:
    if ( key.compareTo(POLY_INSIDE) == 0 )
    {
      insideRule = intProperty( value, insideRule, POLY_INSIDE_PARITY, POLY_INSIDE_WINDING );
      return;
    }
      // SUPERSAMPLING:
    if ( key.compareTo(SUPERSAMPLING) == 0 )
    {
      supersampling = intProperty( value, supersampling, 2, 32 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(POLY_INSIDE) == 0 )
      return insideRule;
    if ( key.compareTo(SUPERSAMPLING) == 0 )
      return supersampling;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "PolygonFillScanline";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME  = "PolygonFillRenderToBitMask";

  /** Object template identifier. */
  protected final static String TEMPLATE2_NAME = "PolygonFillRenderToAlphaMask";

  /** Object description. */
  private final static String DESCRIPTION  = "2D polygon rasterization (w/o anti-aliasing).";

  /** Object description. */
  private final static String DESCRIPTION2 = "2D polygon rasterization with anti-aliasing.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_POLYGON + '.' + C_FLOAT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 1 )
    {
      t.newInputPlug( PL_INPUT, IFACE+"PolygonFillRender" );
        // Plugs:
      t.setRegStrings( NAME, (ord <= 0) ? TEMPLATE_NAME : TEMPLATE2_NAME, CATEGORY,
                       (ord <= 0) ? DESCRIPTION : DESCRIPTION2 );
      if ( ord != 1 ) t.newOutputPlug( PL_BITMASK,   IFACE+"BitMask" );
      if ( ord != 0 ) t.newOutputPlug( PL_ALPHAMASK, IFACE+"AlphaMask" );
        // Property:
      t.propBegin( POLY_INSIDE, TYPE_INTEGER, "Inside-test mode", true );
        t.propDefault( POLY_INSIDE_PARITY );
        t.propBounds( POLY_INSIDE_PARITY, POLY_INSIDE_WINDING );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Parity",  POLY_INSIDE_PARITY,  "Odd-even rule" );
        t.propEnum( "Winding", POLY_INSIDE_WINDING, "Positive winding rule" );
      t.propEnd();
      if ( ord == 1 )
      {
        t.propBegin( SUPERSAMPLING, TYPE_INTEGER, "Supersampling factor", true );
          t.propDefault( 4 );
          t.propBounds( 2, 32 );
        t.propEnd();
      }
    }
    return 2;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece[] reg;
  static
  {
    reg = new RegPiece[ 2 ];
    reg[0] = new RegPiece();
    setTemplate( reg[0], 0 );
    reg[1] = new RegPiece();
    setTemplate( reg[1], 1 );
  }

}
