package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * {@link Brep} scene rendering - depth-buffer implementation.
 *
 * @version 0.26 $Rev: 843 $ $Date: 2009-01-12 12:14:46 +0100 (Mon, 12 Jan 2009) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RenderDepthBuffer.java">RenderDepthBuffer.java</a>
 */
public class RenderDepthBuffer extends DefaultRender3D
{
  //==========================================================================
  //  Data:

  /** Canvas background. */
  protected final static int[] BACKGROUND = {   0,  20,  60, 255 };

  /** Infinite value used in depth-buffer reset. */
  protected float infinity = -1.0f;

  /** Depth-buffer, addressing: {@code buffer[ x + y * bufferWidth ]}. */
  protected float[] buffer  = null;

  /** Buffer width in pixels. */
  protected int bufferWidth  = 0;

  /** Buffer height in pixels. */
  protected int bufferHeight = 0;

  /** Use perspective-correct interpolation? */
  protected boolean correctInterpolation = true;

  //==========================================================================
  //  Support:

  /** Depth-buffer [re]-initialization. */
  protected void initBuffer ( int width, int height )
  {
    if ( buffer == null ||
         width  != bufferWidth ||
         height != bufferHeight )
    {
      bufferWidth  = width;
      bufferHeight = height;
      buffer = new float[ width * height ];
    }

    Arrays.fill( buffer, infinity );
  }

  //==========================================================================
  //  Support classes:

  /** Abstract H-line renderer. */
  protected static interface HLineRenderer
  {
    /** Render horizontal segment (H-line) from {@code [e1.x,y]} to {@code [e2.x,y]}. */
    void renderHLine ( int y, PolyEdge e1, PolyEdge e2 );
  }

  /**
   * Polygon edge.
   */
  protected static class PolyEdge implements Comparable< PolyEdge >
  {
    /** Starting Y-coordinate. */
    public int y0;

    /** X-coordinate of the edge-hline intersection. */
    public double x;

    /** X difference between two adjacent hlines. */
    public double dx;

    /** Z-coordinate of the edge-hline intersection. */
    public double z;

    /** Z difference between two adjacent hlines. */
    public double dz;

    /** Texture U-coordinate of the edge-hline intersection. */
    public double u;

    /** U difference between two adjacent hlines. */
    public double du;

    /** Texture V-coordinate of the edge-hline intersection. */
    public double v;

    /** V difference between two adjacent hlines. */
    public double dv;

    /** Number of hlines to draw. */
    public int count;

    /** Handle of the top vertex. */
    public int topV;

    /** Handle of the bottom vertex. */
    public int bottomV;

    /** Common constructor for flat shader and texture shader. */
    public PolyEdge ( double[] p1, double[] tex1, double[] p2, double[] tex2, boolean correctInterpolation )
    {
      double yy0, dy, yCorr;

      if ( p1[1] < p2[1] )                // top-down edge
      {
        dy    = p2[1] - p1[1];
        yy0   = Math.ceil( p1[1] );
        yCorr = yy0 - p1[1];
        y0    = Formula.round( yy0 );
        dx    = (p2[0] - p1[0]) / dy;
        x     = p1[0] + yCorr * dx;
        dz    = (p2[2] - p1[2]) / dy;
        z     = p1[2] + yCorr * dz;

        if ( tex1 != null &&
             tex2 != null )
          if ( correctInterpolation )
          {
            du  = (tex2[0] * (1.0 - p2[2]) - tex1[0] * (1.0 - p1[2])) / dy;
            u   = tex1[0] * (1.0 - p1[2]) + yCorr * du;
            dv  = (tex2[1] * (1.0 - p2[2]) - tex1[1] * (1.0 - p1[2])) / dy;
            v   = tex1[1] * (1.0 - p1[2]) + yCorr * dv;
          }
          else
          {
            du  = (tex2[0] - tex1[0]) / dy;
            u   = tex1[0] + yCorr * du;
            dv  = (tex2[1] - tex1[1]) / dy;
            v   = tex1[1] + yCorr * dv;
          }

        count = Formula.round( Math.ceil( p2[1] ) - yy0 );
      }

      else                                // bottom-up edge
      {
        dy    = p1[1] - p2[1];
        yy0   = Math.ceil( p2[1] );
        yCorr = yy0 - p2[1];
        y0    = Formula.round( yy0 );
        dx    = (p1[0] - p2[0]) / dy;
        x     = p2[0] + yCorr * dx;
        dz    = (p1[2] - p2[2]) / dy;
        z     = p2[2] + yCorr * dz;

        if ( tex1 != null &&
             tex2 != null )
          if ( correctInterpolation )
          {
            du  = (tex1[0] * (1.0 - p1[2]) - tex2[0] * (1.0 - p2[2])) / dy;
            u   = tex2[0] * (1.0 - p2[2]) + yCorr * du;
            dv  = (tex1[1] * (1.0 - p1[2]) - tex2[1] * (1.0 - p2[2])) / dy;
            v   = tex2[1] * (1.0 - p2[2]) + yCorr * dv;
          }
          else
          {
            du  = (tex1[0] - tex2[0]) / dy;
            u   = tex2[0] + yCorr * du;
            dv  = (tex1[1] - tex2[1]) / dy;
            v   = tex2[1] + yCorr * dv;
          }

        count = Formula.round( Math.ceil( p1[1] ) - yy0 );
      }
    }

    /** Common constructor for Gouraud/Phong shader and every other interpolating shader. */
    public PolyEdge ( double[] p1, int h1, double[] p2, int h2, boolean correctInterpolation )
    {
      double yy0, dy, yCorr;

      if ( p1[1] < p2[1] )                // top-down edge
      {
        dy    = p2[1] - p1[1];
        yy0   = Math.ceil( p1[1] );
        yCorr = yy0 - p1[1];
        y0    = Formula.round( yy0 );
        dx    = (p2[0] - p1[0]) / dy;
        x     = p1[0] + yCorr * dx;
        dz    = (p2[2] - p1[2]) / dy;
        z     = p1[2] + yCorr * dz;
        topV  = h1;
        bottomV = h2;

        du    = (correctInterpolation ? (1.0 - p2[2]) : 1.0) / dy;
        u     = yCorr * du;

        count = Formula.round( Math.ceil( p2[1] ) - yy0 );
      }

      else                                // bottom-up edge
      {
        dy    = p1[1] - p2[1];
        yy0   = Math.ceil( p2[1] );
        yCorr = yy0 - p2[1];
        y0    = Formula.round( yy0 );
        dx    = (p1[0] - p2[0]) / dy;
        x     = p2[0] + yCorr * dx;
        dz    = (p1[2] - p2[2]) / dy;
        z     = p2[2] + yCorr * dz;
        topV  = h2;
        bottomV = h1;

        du    = (correctInterpolation ? (1.0 - p1[2]) : 1.0) / dy;
        u     = yCorr * du;

        count = Formula.round( Math.ceil( p1[1] ) - yy0 );
      }
    }

    /** Prepare the edge record for the next line. */
    public void next ()
    {
      x    += dx;
      z    += dz;
      u    += du;
      v    += dv;
    }

    /** Natural ordering: [ y0, x ]. */
    public int compareTo ( PolyEdge o )
    {
      if ( o == null )
        throw new NullPointerException();

      if ( this == o )
        return 0;

      if ( y0 != o.y0 )
        return( y0 < o.y0 ? -1 : 1 );

      if ( x != o.x )
        return( x < o.x ? -1 : 1 );

      return( hashCode() < o.hashCode() ? -1 : 1 );
    }
  }

  /** Edge comparator for actual edge list: [ x ]. */
  protected static class EdgeComparator implements Comparator< PolyEdge >
  {
    public int compare ( PolyEdge a, PolyEdge b )
    {
      if ( a == b )
        return 0;

      if ( a.x != b.x )
        return( a.x < b.x ? -1 : 1 );

      return( a.hashCode() < b.hashCode() ? -1 : 1 );
    }
  }

  /** Single comparator instance. */
  protected final static EdgeComparator eComparator = new EdgeComparator();

  /** H-line renderer - flat shading. */
  protected class FlatShader implements HLineRenderer
  {
    protected final RasterGraphics output;

    public FlatShader ( RasterGraphics out )
    {
      output = out;
    }

    /** Render horizontal segment (H-line) from {@code [ e1.x, y ]} to {@code [ e2.x, y ]}. */
    public void renderHLine ( int y, PolyEdge e1, PolyEdge e2 )
    {
      if ( y < 0 || y >= bufferHeight ) return;

      int x1 = Formula.round( Math.ceil( e1.x ) );
      int x2 = Formula.round( Math.ceil( e2.x ) );

      if ( x1 >= x2 ||
           x1 >= bufferWidth ||
           x2 <= 0 ) return;

      final double dZ = (e2.z - e1.z) / (e2.x - e1.x);
      double z        = e1.z + dZ * (x1 - e1.x);
      int ind         = y * bufferWidth + x1;
      if ( x2 > bufferWidth ) x2 = bufferWidth;

      switch ( depthComparison )
      {
        case DEPTH_GREATER:
          while ( x1 < x2 )
          {
            if ( x1 >= 0 &&
                 z  > buffer[ ind ] )
            {
              buffer[ ind ] = (float)z;
              output.setPixel( x1, y );
            }
            z += dZ;
            x1++;
            ind++;
          }
          break;

        case DEPTH_GEQUAL:
          while ( x1 < x2 )
          {
            if ( x1 >= 0 &&
                 z  >= buffer[ ind ] )
            {
              buffer[ ind ] = (float)z;
              output.setPixel( x1, y );
            }
            z += dZ;
            x1++;
            ind++;
          }
          break;

        case DEPTH_LEQUAL:
          while ( x1 < x2 )
          {
            if ( x1 >= 0 &&
                 z  <= buffer[ ind ] )
            {
              buffer[ ind ] = (float)z;
              output.setPixel( x1, y );
            }
            z += dZ;
            x1++;
            ind++;
          }
          break;

        case DEPTH_LESS:
          while ( x1 < x2 )
          {
            if ( x1 >= 0 &&
                 z  < buffer[ ind ] )
            {
              buffer[ ind ] = (float)z;
              output.setPixel( x1, y );
            }
            z += dZ;
            x1++;
            ind++;
          }
          break;
      }
    }
  }

  /**
   * H-line renderer - external texture mapping
   * (with optional perspective-correct interpolation).
   */
  protected class TextureShader implements HLineRenderer
  {
    protected final RasterGraphics output;

    protected final ImageFunction texture;

    protected final double[] color = new double[ 3 ];

    public TextureShader ( RasterGraphics out, ImageFunction tex )
    {
      output  = out;
      texture = tex;
      texture.setBounds( 0.0, 1.0, 0.0, 1.0 );
    }

    /** Render horizontal segment (H-line) from {@code [ e1.x, y ]} to {@code [ e2.x, y ]}. */
    public void renderHLine ( int y, PolyEdge e1, PolyEdge e2 )
    {
      if ( y < 0 || y >= bufferHeight ) return;

      int x1 = Formula.round( Math.ceil( e1.x ) );
      int x2 = Formula.round( Math.ceil( e2.x ) );

      if ( x1 >= x2 ||
           x1 >= bufferWidth ||
           x2 <= 0 ) return;

      final double dX = e2.x - e1.x;
      final double cX = x1 - e1.x;
      final double dZ = (e2.z - e1.z) / dX;
      double z        = e1.z + dZ * cX;
      final double dU = (e2.u - e1.u) / dX;
      double u        = e1.u + dU * cX;
      final double dV = (e2.v - e1.v) / dX;
      double v        = e1.v + dV * cX;

      int ind         = y * bufferWidth + x1;
      if ( x2 > bufferWidth ) x2 = bufferWidth;

      switch ( depthComparison )
      {
        case DEPTH_GREATER:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  > buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u / (1.0 - z), v / (1.0 - z), color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  > buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u, v, color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          break;

        case DEPTH_GEQUAL:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  >= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u / (1.0 - z), v / (1.0 - z), color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  >= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u, v, color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          break;

        case DEPTH_LEQUAL:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  <= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u / (1.0 - z), v / (1.0 - z), color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  <= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u, v, color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          break;

        case DEPTH_LESS:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  < buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u / (1.0 - z), v / (1.0 - z), color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  < buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;
                texture.getSample( u, v, color );
                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              v += dV;
              x1++;
              ind++;
            }
          break;
      }
    }
  }

  /**
   * Interpolating renderer - Gouraud color interpolation
   * (with optional perspective-correct interpolation).
   */
  protected class GouraudShader implements HLineRenderer
  {
    protected final RasterGraphics output;

    protected final Brep brep;

    /** Handle of vertex-color attribute in the {@link #brep}. */
    protected final int colorAttr;

    protected final float[]  colL  = new float[ 3 ];
    protected final float[]  colR  = new float[ 3 ];
    protected final float[]  colA  = new float[ 3 ];
    /** Accelerator: output color. */
    protected final double[] color = new double[ 3 ];

    public GouraudShader ( RasterGraphics out, Brep br, int attr )
    {
      output    = out;
      brep      = br;
      colorAttr = attr;
    }

    /** Render horizontal segment (H-line) from {@code [ e1.x, y ]} to {@code [ e2.x, y ]}. */
    public void renderHLine ( int y, PolyEdge e1, PolyEdge e2 )
    {
      if ( y < 0 || y >= bufferHeight ) return;

      int x1 = Formula.round( Math.ceil( e1.x ) );
      int x2 = Formula.round( Math.ceil( e2.x ) );

      if ( x1 >= x2 ||
           x1 >= bufferWidth ||
           x2 <= 0 ) return;

      final double dX = e2.x - e1.x;
      final double cX = x1 - e1.x;
      final double dZ = (e2.z - e1.z) / dX;
      double z        = e1.z + dZ * cX;
      final double dU = (correctInterpolation ? (1.0 - e2.z) : 1.0) / dX;
      double u        = dU * cX;
      float coef;
      double coefd;

      int ind         = y * bufferWidth + x1;
      if ( x2 > bufferWidth ) x2 = bufferWidth;

        // (1.0 - u) * ((1.0 -  left.u) *  left.top +  left.u *  left.bottom) +
        //        u  * ((1.0 - right.u) * right.top + right.u * right.bottom)
        // .. replace "u" by "u/z" for perspective-correct interpolation
      brep.getAttribute( colorAttr, e1.bottomV, colL ); // left.bottom color
      coef     = (float)(correctInterpolation ? e1.u / (1.0 - e1.z) : e1.u);
      colL[0] *= coef;
      colL[1] *= coef;
      colL[2] *= coef;
      brep.getAttribute( colorAttr, e1.topV, colA );    // left.top color
      coef     = 1.0f - coef;
      colL[0] += coef * colA[0];
      colL[1] += coef * colA[1];
      colL[2] += coef * colA[2];
      brep.getAttribute( colorAttr, e2.bottomV, colR ); // right.bottom color
      coef     = (float)(correctInterpolation ? e2.u / (1.0 - e2.z) : e2.u);
      colR[0] *= coef;
      colR[1] *= coef;
      colR[2] *= coef;
      brep.getAttribute( colorAttr, e2.topV, colA );    // right.top color
      coef     = 1.0f - coef;
      colR[0] += coef * colA[0];
      colR[1] += coef * colA[1];
      colR[2] += coef * colA[2];

      switch ( depthComparison )
      {
        case DEPTH_GREATER:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  > buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u/z) * left + u/z  * right
                coefd    = u / (1.0 - z);
                color[0] = (1.0 - coefd) * colL[0] + coefd * colR[0];
                color[1] = (1.0 - coefd) * colL[1] + coefd * colR[1];
                color[2] = (1.0 - coefd) * colL[2] + coefd * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  > buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u) * left + u  * right
                coefd    = 1.0 - u;
                color[0] = coefd * colL[0] + u * colR[0];
                color[1] = coefd * colL[1] + u * colR[1];
                color[2] = coefd * colL[2] + u * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          break;

        case DEPTH_GEQUAL:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  >= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u/z) * left + u/z  * right
                coefd    = u / (1.0 - z);
                color[0] = (1.0 - coefd) * colL[0] + coefd * colR[0];
                color[1] = (1.0 - coefd) * colL[1] + coefd * colR[1];
                color[2] = (1.0 - coefd) * colL[2] + coefd * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  >= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u) * left + u  * right
                coefd    = 1.0 - u;
                color[0] = coefd * colL[0] + u * colR[0];
                color[1] = coefd * colL[1] + u * colR[1];
                color[2] = coefd * colL[2] + u * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          break;

        case DEPTH_LEQUAL:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  <= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u/z) * left + u/z  * right
                coefd    = u / (1.0 - z);
                color[0] = (1.0 - coefd) * colL[0] + coefd * colR[0];
                color[1] = (1.0 - coefd) * colL[1] + coefd * colR[1];
                color[2] = (1.0 - coefd) * colL[2] + coefd * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  <= buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u) * left + u  * right
                coefd    = 1.0 - u;
                color[0] = coefd * colL[0] + u * colR[0];
                color[1] = coefd * colL[1] + u * colR[1];
                color[2] = coefd * colL[2] + u * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          break;

        case DEPTH_LESS:
          if ( correctInterpolation )
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  < buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u/z) * left + u/z  * right
                coefd    = u / (1.0 - z);
                color[0] = (1.0 - coefd) * colL[0] + coefd * colR[0];
                color[1] = (1.0 - coefd) * colL[1] + coefd * colR[1];
                color[2] = (1.0 - coefd) * colL[2] + coefd * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          else
            while ( x1 < x2 )
            {
              if ( x1 >= 0 &&
                   z  < buffer[ ind ] )
              {
                buffer[ ind ] = (float)z;

                  // (1.0 - u) * left + u  * right
                coefd    = 1.0 - u;
                color[0] = coefd * colL[0] + u * colR[0];
                color[1] = coefd * colL[1] + u * colR[1];
                color[2] = coefd * colL[2] + u * colR[2];

                output.putPixel( x1, y, color );
              }
              z += dZ;
              u += dU;
              x1++;
              ind++;
            }
          break;
      }
    }
  }

  //==========================================================================
  //  Interface Render3D:

  /** Sets type of depth-test comparison. */
  @Override
  public void setDepthComparison ( int comp )
  {
    switch ( depthComparison = comp )
    {
      case DEPTH_LESS:
      case DEPTH_LEQUAL:
        infinity = Float.POSITIVE_INFINITY;
        break;

      case DEPTH_GEQUAL:
      case DEPTH_GREATER:
        infinity = Float.NEGATIVE_INFINITY;
    }
  }

  /**
   * 3D rendering using the given projection matrix.
   *
   * @param mv Local model-view matrix.
   * @param proj Local projection matrix.
   * @param style Local rendering style.
   * @see #render()
   */
  @Override
  public void render ( TrMatrix mv, TrMatrix proj, int style )
  {
    final Brep           data   = (Brep)          getInterface( PL_DATA,   IFACE+"Brep" );
    final RasterGraphics output = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    final ImageFunction  texIm  = (ImageFunction) getInterface( PL_IMAGE,  IFACE+"ImageFunction" );

    if ( data   == null ||
         output == null ) return;

      // projection:
    setModelView( mv );
    setProjection( proj );
    checkMatrix();

    final int width   = output.getWidth();
    final int height  = output.getHeight();
    output.setRectangle( 0, 0, width, height, BACKGROUND );
    initBuffer( width, height );

    final int ctx     = Brep.CTX_DEFAULT;
    final int worldId = data.getAttributeId( Brep.VERTEX, ctx, Brep.COORD );
    final int projId  = data.transformCoordinates( worldId, ctx, Brep.COORD_PROJ, compound );
    final int colorId = data.getAttributeId( Brep.FACE, ctx, Brep.COLOR );
    final int normalId = data.getAttributeId( Brep.VERTEX, ctx, Brep.NORMAL );
    final int texId   = data.getAttributeId( Brep.VERTEX, ctx, Brep.TEXTURE_0 );
    final int vColId  = data.getAttributeId( Brep.VERTEX, ctx, Brep.COLOR );

    BrepIterator it   = data.faceIterator( ctx, null );
    int faceId;
    int[] vert  = new int[ 3 ];
    int faceSize;
    int i, i1;
    double[][] vc  = new double[ 2 ][ 4 ];
    float[] colorF = new float[ 4 ];
    double[] color = new double[ 3 ];
    double[] norm  = new double[ 4 ];
    double[] world = new double[ 4 ];
    double[][] tex = new double[ 2 ][ 2 ];

      // input edge set sorted by [ y0 ]
    TreeSet< PolyEdge > input  = new TreeSet< PolyEdge >();
      // actual edge list sorted by [ x ]
    TreeSet< PolyEdge > actual = new TreeSet< PolyEdge >( eComparator );
    TreeSet< PolyEdge > backup = new TreeSet< PolyEdge >( eComparator );

      // H-line renderer:
    final HLineRenderer rend;
    switch ( renderStyle )
    {
      case RENDER_STYLE_GOURAUD:
        rend = new GouraudShader( output, data, vColId );
        break;

      case RENDER_STYLE_TEXTURE:
        if ( texIm != null )
        {
          rend = new TextureShader( output, texIm );
          break;
        }

      default:
        rend = new FlatShader( output );
    }

    while ( (faceId = it.next()) != Brep.NULL &&
            !userBreak )
    {
        // 1. initialize all the lists:
      input.clear();
      actual.clear();
      backup.clear();

        // 2. sort input edge records into input list
      while ( (faceSize = data.getFaceVertices( faceId, vert )) > vert.length )
        vert = new int[ faceSize ];

      data.getAttribute( colorId, faceId, colorF );
      color[0] = colorF[0];
      color[1] = colorF[1];
      color[2] = colorF[2];
      output.setColor( color );

      int newV = vert[0];
      int oldV;
      data.getAttribute( projId, newV, vc[1] );
      Geometry.toCartesian3D( vc[1], vc[1] );
      data.getAttribute( texId,  newV, tex[1] );

      for ( i = 0; i < faceSize; i++ )
      {
        if ( (i1 = i + 1) >= faceSize ) i1 = 0;

        oldV = newV;
        newV = vert[i1];
        data.getAttribute( projId, newV, vc[i & 1] );
        Geometry.toCartesian3D( vc[i & 1], vc[i & 1] );
        data.getAttribute( texId,  newV, tex[i & 1] );

          // include edge vc[0]--vc[1]:
        if ( Math.ceil( vc[0][1] ) == Math.ceil( vc[1][1] ) )
          continue;                       // discarding horizontal edges

        if ( renderStyle == RENDER_STYLE_GOURAUD )
          input.add( new PolyEdge( vc[(i-1) & 1], oldV, vc[i & 1], newV, correctInterpolation ) );
        else
          input.add( new PolyEdge( vc[0], tex[0], vc[1], tex[1], correctInterpolation ) );
      }

        // 3. initialize actual edge set
      if ( input.size() == 0 ) continue;
      PolyEdge act = input.first();
      input.remove( act );                // !!! TODO(1.6): act = input.pollFirst()

      int y = act.y0;
      actual.add( act );

        // 4. hline loop
      do
      {
          // move all relevant edges from input into actual:
        while ( input.size() > 0 &&
                (act = input.first()).y0 == y )
        {                                 // !!! TODO(1.6): act = input.pollFirst()
          input.remove( act );
          actual.add( act );
        }

        PolyEdge old  = null;             // previous edge (kept in current-line state for potential rendering)
        int      rule = 0;

        do
        {
            // 4a. edge contribution to the current hline
          act  = actual.first();
          actual.remove( act );           // !!! TODO(1.6): act = actual.pollFirst()

          if ( rule != 0 )                // draw H-line segment from the previous edge..
            rend.renderHLine( y, old, act );

            // 4b. old edge ages for the next hline
          if ( old != null &&
               --old.count > 0 )
          {
            old.next();
            backup.add( old );
          }

          rule = 1 - rule;
          old  = act;
        }
        while ( actual.size() > 0 );

          // the last edge was not transformed..
        if ( --old.count > 0 )
        {
          old.next();
          backup.add( old );
        }

          // swap the actual sets:
        TreeSet< PolyEdge > tmp = backup;
        backup = actual;
        actual = tmp;

          // Y-coordinate of the next line
        y++;
      }
      while ( actual.size() > 0 ||
              input.size()  > 0 );
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String CORRECT_INTERPOLATION = "Correct interp.";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // RENDER_STYLE:
    if ( key.compareTo(RENDER_STYLE) == 0 )
    {
      renderStyle = intProperty( value, renderStyle, RENDER_STYLE_LINE, RENDER_STYLE_TEXTURE );
      return;
    }
      // CORRECT_INTERPOLATION:
    if ( key.compareTo(CORRECT_INTERPOLATION) == 0 )
    {
      correctInterpolation = booleanProperty( value, correctInterpolation );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(RENDER_STYLE) == 0 )
      return renderStyle;
    if ( key.compareTo(CORRECT_INTERPOLATION) == 0 )
      return correctInterpolation;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Render depth-buffer";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Render3DToBrepAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Depth-buffer 3D rendering.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_BREP;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 1 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"Render3D" );
      t.newInputPlug(     PL_DIRECT, IFACE+"Trigger" );
      t.newOutputPlug(    PL_DATA,   IFACE+"Brep" );
      t.newOutputPlug(    PL_OUTPUT, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_IMAGE,  IFACE+"ImageFunction" );
        // Property:
      t.propBegin( RENDER_STYLE, TYPE_INTEGER, "Render style", true );
        t.propDefault( RENDER_STYLE_LINE );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Line drawing",    RENDER_STYLE_LINE,    "Simple wire-frame drawing (w/o visibility)" );
        t.propEnum( "Flat shading",    RENDER_STYLE_FLAT,    "Flat-shading (visibility, constant face color)" );
        t.propEnum( "Gouraud shading", RENDER_STYLE_GOURAUD, "Gouraud shading (visibility, colors interpolated from vertices)" );
        t.propEnum( "Phong shading",   RENDER_STYLE_PHONG,   "Phong shading (visibility, normals interpolated from vertices)" );
        t.propEnum( "Texture mapping", RENDER_STYLE_TEXTURE, "Single 2D texture-mapping (visibility, external=procedural texture)" );
      t.propEnd();
      t.propBegin( CORRECT_INTERPOLATION, TYPE_BOOLEAN, "Use perspective-correct interpolation?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
